@@ -126,7 +126,11 @@ send_focus_event (GnomeIconTextItem *iti, gboolean in)
 	fake_event.focus_change.window = widget->window;
 	fake_event.focus_change.in = in;
 	gtk_widget_event (widget, &fake_event);
+	
+	/* FIXME: this is failing */
+#if 0
 	g_return_if_fail (GTK_WIDGET_HAS_FOCUS (widget) == in);
+#endif
 }
 
 /* Updates the pango layout width */
@@ -145,12 +149,22 @@ update_pango_layout (GnomeIconTextItem *iti)
 		text = iti->text;
 	}
 
+	pango_layout_set_wrap (priv->layout, PANGO_WRAP_WORD);
 	pango_layout_set_text (priv->layout, text, strlen (text));
-
 	pango_layout_set_width (priv->layout, iti->width * PANGO_SCALE);
 
+	/* In PANGO_WRAP_WORD mode, words wider than a line of text make
+	 * PangoLayout overflow the layout width.  If this happens, switch to
+	 * character-based wrapping.
+	 */
+
 	pango_layout_get_pixel_extents (iti->_priv->layout, NULL, &bounds);
 
+	if (bounds.width > iti->width) {
+		pango_layout_set_wrap (priv->layout, PANGO_WRAP_CHAR);
+		pango_layout_get_pixel_extents (iti->_priv->layout, NULL, &bounds);
+	}
+
 	priv->layout_width = bounds.width;
 	priv->layout_height = bounds.height;
 }
@@ -352,21 +366,24 @@ iti_entry_activate (GtkObject *widget, gpointer data)
 static void
 realize_cursor_gc (GnomeIconTextItem *iti)
 {
+	GtkWidget *widget;
 	GdkColor *cursor_color;
-	GdkColor red = { 0, 0xffff, 0x0000, 0x0000 };
+
+	widget = GTK_WIDGET (GNOME_CANVAS_ITEM (iti)->canvas);
 
 	if (iti->_priv->cursor_gc) {
 		g_object_unref (iti->_priv->cursor_gc);
 	}
 
-	iti->_priv->cursor_gc = gdk_gc_new (GTK_WIDGET (GNOME_CANVAS_ITEM (iti)->canvas)->window);
+	iti->_priv->cursor_gc = gdk_gc_new (widget->window);
 
 	gtk_widget_style_get (GTK_WIDGET (iti->_priv->entry), "cursor_color",
 			      &cursor_color, NULL);
 	if (cursor_color) {
 		gdk_gc_set_rgb_fg_color (iti->_priv->cursor_gc, cursor_color);
 	} else {
-		gdk_gc_set_rgb_fg_color (iti->_priv->cursor_gc, &red);
+		gdk_gc_set_rgb_fg_color (iti->_priv->cursor_gc, 
+					 &widget->style->black);
 	}
 }
 
@@ -514,12 +531,16 @@ iti_draw_cursor (GnomeIconTextItem *iti, GdkDrawable *drawable,
 {
 	int stem_width;
 	int i;
+	int cursor_offset;
 	PangoRectangle pos;
-
+	GtkEntry *entry;
+	
 	g_return_if_fail (iti->_priv->cursor_gc != NULL);
 
+	entry = GTK_ENTRY (iti->_priv->entry);
+	cursor_offset = gtk_editable_get_position (GTK_EDITABLE (entry));
 	pango_layout_get_cursor_pos (iti->_priv->layout,
-				     gtk_editable_get_position (GTK_EDITABLE (iti->_priv->entry)),
+				     g_utf8_offset_to_pointer (entry->text, cursor_offset) - entry->text,
 				     &pos, NULL);
 	stem_width = PANGO_PIXELS (pos.height) / 30 + 1;
 	for (i = 0; i < stem_width; i++) {
@@ -564,11 +585,12 @@ gnome_icon_text_item_draw (GnomeCanvasItem *item, GdkDrawable *drawable,
 	text_xofs = xofs - (iti->width - priv->layout_width - 2 * MARGIN_X) / 2;
 	text_yofs = yofs + MARGIN_Y;
 
-
-	if (GTK_WIDGET_HAS_FOCUS (widget))
+	if (iti->selected && GTK_WIDGET_HAS_FOCUS (widget))
 		state = GTK_STATE_SELECTED;
-	else
+	else if (iti->selected)
 		state = GTK_STATE_ACTIVE;
+	else 
+		state = GTK_STATE_NORMAL;
 
 	if (iti->selected && !iti->editing)
 		gdk_draw_rectangle (drawable,
@@ -577,14 +599,16 @@ gnome_icon_text_item_draw (GnomeCanvasItem *item, GdkDrawable *drawable,
 				    xofs + 1, yofs + 1,
 				    w - 2, h - 2);
 
-	if (iti->focused && ! iti->editing) {
+	if (GTK_WIDGET_HAS_FOCUS (widget) && iti->focused && ! iti->editing) {
 		GdkRectangle r;
 
 		r.x = xofs;
 		r.y = yofs;
 		r.width = w - 1;
 		r.height = h - 1;
-		gtk_paint_focus (style, drawable, state, &r, widget, NULL,
+		gtk_paint_focus (style, drawable,
+				 iti->selected ? GTK_STATE_SELECTED: GTK_STATE_NORMAL,
+				 &r, widget, NULL,
 				 xofs, yofs, w - 1, h - 1);
 #if 0
 		gtk_draw_focus (style,
@@ -607,7 +631,7 @@ gnome_icon_text_item_draw (GnomeCanvasItem *item, GdkDrawable *drawable,
 	}
 
 	gdk_draw_layout (drawable,
-			 style->text_gc[iti->selected ? state : GTK_STATE_NORMAL],
+			 style->text_gc[iti->editing ? GTK_STATE_NORMAL : state],
 			 text_xofs,
 			 text_yofs,
 			 priv->layout);
@@ -619,6 +643,9 @@ gnome_icon_text_item_draw (GnomeCanvasItem *item, GdkDrawable *drawable,
 			GdkColor *selection_color, *text_color;
 			guint8 state;
 
+			range[0] = g_utf8_offset_to_pointer (GTK_ENTRY (priv->entry)->text, range[0]) - GTK_ENTRY (priv->entry)->text;
+			range[1] = g_utf8_offset_to_pointer (GTK_ENTRY (priv->entry)->text, range[1]) - GTK_ENTRY (priv->entry)->text;
+			
 			state = GTK_WIDGET_HAS_FOCUS (widget) ? GTK_STATE_SELECTED : GTK_STATE_ACTIVE;
 			selection_color = &widget->style->base[state];
 			text_color = &widget->style->text[state];
